/******************************************************************************
 * sampler.c - Copyright (c) 2012 Dario Ghersi                                *
 * Version: 20120724                                                          *
 * Goal:    perform a Monte Carlo sampling of genes to achieve a distribution *
 *          of GO terms that matches as closely as ossible a target          *
 *                                                                            *
 * License: This file is part of SAMPLER.                                     *
 *          SAMPLER is free software: you can redistribute it and/or modify   *
 *          it under the terms of the GNU General Public License as           *
 *          published by the Free Software Foundation, either version 3 of    *
 *          the License, or (at your option) any later version.               *
 *
 *          SAMPLER is distributed in the hope that it will be useful,        *
 *          but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *          MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
 *          GNU General Public License for more details.                      *
 *                                                                            *
 *          You should have received a copy of the GNU General Public License *
 *          along with SAMPLER.  If not, see <http://www.gnu.org/licenses/>.  *
 ******************************************************************************/

#include <gsl/gsl_randist.h>
#include <gsl/gsl_rng.h>
#include <math.h>
#include "sampler.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

/******************************************************************************
 * FUNCTIONS DEFINITIONS                                                      *
 ******************************************************************************/

double calculate_dist(target_distr target, unsigned int *term_distr)
{
  /* calculate the squared euclidean norm between the target distribution
     and the distribution obtained from the current sample */

  double squared_norm = 0.0;
  unsigned int i;

  for (i = 0; i < target.num_terms; i++) {
    squared_norm += (target.frequency[i] - term_distr[i]) *
      (target.frequency[i] - term_distr[i]);
  }  

  return squared_norm;
}

/******************************************************************************/

void calculate_term_distr(unsigned int *term_distr, unsigned int *sample,
			  unsigned int size_sample, annotation *ann)
{
  /* calculate the term distribution of the sample */

  unsigned int i, j, geneID;

  for (i = 0; i < size_sample; i++) {
    geneID = sample[i];
    for (j = 0; j < ann[geneID].num_terms; j++) {
      term_distr[ann[geneID].terms[j]]++;
    }
  }

}

/******************************************************************************/

sampling_log sampler(target_distr target, annotation *ann,
		     unsigned int *control_genes,
		     const unsigned int size_control,
		     const unsigned int size_sample, double max_iter,
		     double tolerance, unsigned int *sample, double exp_coeff)
{
  /* perform Markov Chain Monte Carlo sampling */
  unsigned int i, accept;
  double sq_norm = 1E12, current_norm, p, r;
  unsigned int *term_distr, *current_term_distr, sample_pos, count = 0;
  unsigned int *control_avail, size_control_avail, control_pos, temp;
  sampling_log log;
 
  /* randomize the control genes */
  gsl_ran_shuffle(rng, control_genes, size_control, sizeof (unsigned int));

  /* initialize the sample */
  for (i = 0; i < size_sample; i++) {
    sample[i] = control_genes[i];
  }

  /* create a CONTROL_GENES - SAMPLE set */
  size_control_avail = size_control - size_sample;
  control_avail = malloc(sizeof(unsigned int) * size_control_avail);
  count = 0;
  for (i = size_sample; i < size_control; i++) {
    control_avail[count++] = control_genes[i];
  }

  /* calculate the term distribution in the sample */
  term_distr = malloc(sizeof(unsigned int) * target.num_terms);
  current_term_distr = malloc(sizeof(unsigned int) * target.num_terms);
  for (i = 0; i < target.num_terms; i++) {
    term_distr[i] = 0;
  }
  calculate_term_distr(term_distr, sample, size_sample, ann);
  sq_norm = current_norm = calculate_dist(target, term_distr);

  /* Monte Carlo step */
  count = 0;
  while(count < max_iter && sq_norm > tolerance) {
    /* select a position to swap in the sample */
    sample_pos = gsl_rng_uniform_int(rng, size_sample);

    /* select a position to swap in the control pool */
    control_pos = gsl_rng_uniform_int(rng, size_control_avail);

    /* recalculate the term distribution with the swap */
    for (i = 0; i < target.num_terms; i++) {
      current_term_distr[i] = term_distr[i];
    }
    recalculate_term_distr(current_term_distr, ann,
			   sample[sample_pos], control_avail[control_pos]);
    current_norm = calculate_dist(target, current_term_distr);

    /* accept or reject the move */
    if (exp_coeff > 0) { /* Metropolis criterion */
      p = min(1.0, exp(exp_coeff * (sq_norm - current_norm)));
      r = gsl_rng_uniform_pos(rng);
 
      if (r <= p) {
        accept = 1;
      }
      else {
	accept = 0;
      }
    }
    else { /* no Metropolis criterion */
      accept = (current_norm <= sq_norm);
    }

    if (accept) { /* swap */
      sq_norm = current_norm;
      temp = sample[sample_pos];
      sample[sample_pos] = control_avail[control_pos];
      control_avail[control_pos] = temp;

      /* update the term distribution */
      for (i = 0; i < target.num_terms; i++) {
        term_distr[i] = current_term_distr[i];
      }
    }
    count++;
  }
  
  /* fill in the log */
  log.squared_norm = sq_norm;
  log.num_iter = count;

  free(term_distr);
  free(control_avail);

  return log;
}

/******************************************************************************/

void print_samples(FILE *out_file, unsigned int *sample,
		   unsigned int size_sample, annotation *ann)
{
  /* print the samples in a separate line */

  unsigned int i;

  fprintf(out_file, "%s", ann[sample[0]].gene_name);
  for (i = 1; i < size_sample; i++) {
    fprintf(out_file, " %s", ann[sample[i]].gene_name);
  }
  fprintf(out_file, "\n");

}

/******************************************************************************/

int read_annotations(char *ann_file_name, annotation **ann, target_distr target)
{
  /* store the annotations into an array (genes) of arrays (annotations) */

  FILE *ann_file;
  unsigned int i = 0, j, num_annotations = 0, num_terms = 0, count = 0;
  char *line, *buf, *newline;

  /* open the file and read how many lines it has */
  ann_file = fopen(ann_file_name, "r");
  line = malloc(sizeof(char) * READ_BUF);
  newline = malloc(sizeof(char) * READ_BUF);
  while (fgets(line, READ_BUF, ann_file) != NULL) {
    num_annotations++;
  }

  /* allocate memory for ann */
  (*ann) = malloc(sizeof(annotation) * num_annotations);

  /* store the annotations */
  rewind(ann_file);
  i = 0;
  while (fgets(line, READ_BUF, ann_file) != NULL) {
    (*ann)[i].num_terms = 0;
    strncpy(newline, line, READ_BUF);

    /* assign the gene name */
    buf = strtok(line, " \t");
    strncpy((*ann)[i].gene_name, buf, GENE_NAME_SIZE);

    /* determine how many annotations the gene has */
    while((buf = strtok(NULL, " \t"))) {
      (*ann)[i].num_terms++;
    }
    num_terms = (*ann)[i].num_terms;

    /* store the annotations */
    (*ann)[i].terms = malloc(sizeof(unsigned int) * num_terms);
    count = 0;

    /* remove the newline character */
    newline[strlen(newline) - 1] = '\0';

    /* tokenize the line to get the terms */
    buf = strtok(newline, " \t");
    while ((buf = strtok(NULL, " \t"))) {
      for (j = 0; j < target.num_terms; j++) {
	if (strcmp(buf, target.terms[j]) == 0) { /* term found */
	  (*ann)[i].terms[count] = j;
	  break;
	}
      }
      count++;
    }

    i++;
  }

  free(line);
  free(newline);
  fclose(ann_file);

  return num_annotations;
}

/******************************************************************************/

void read_target_distr(char *target_file_name, target_distr *target)
{
  /* store the target distribution into a structure with terms and frequency */

  FILE *target_file;
  unsigned int i = 0;
  char *line, *buf;

  /* open the file and determine how many lines it has */
  target_file = fopen(target_file_name, "r");
  target->num_terms = 0;
  line = malloc(sizeof(char) * READ_BUF);
  while (fgets(line, READ_BUF, target_file) != NULL) {
    target->num_terms++;
  }

  /* allocate the memory for the target_distr */
  target->terms = malloc(sizeof(char *) * target->num_terms);
  for (i = 0; i < target->num_terms; i++) {
    target->terms[i] = malloc(sizeof(char) * GO_SIZE);
  }
  target->frequency = malloc(sizeof(unsigned int) * target->num_terms);

  /* store the go terms and the target frequency */
  rewind(target_file);
  i = 0;
  
  while (fgets(line, READ_BUF, target_file) != NULL) {
    /* get the GO term */
    buf = strtok(line, " \t");
    strncpy(target->terms[i], buf, GO_SIZE);

    /* get the frequency */
    buf = strtok(NULL, " \t");
    target->frequency[i] = atoi(buf);
    i++;
  }

  free(line);
  fclose(target_file);
}

/******************************************************************************/

double read_control_genes(char *control_genes_file_name, annotation *ann,
			       unsigned int num_ann,
			       unsigned int **control_genes)
{
  /* assign to the control genes their position in the annotation array */

  FILE *control_genes_file;
  char *line;
  unsigned int i, count = 0, num_genes = 0;

  /* open the file and determine how many lines it has */
  control_genes_file = fopen(control_genes_file_name, "r");
  line = malloc(sizeof(char) * READ_BUF);
  while (fgets(line, READ_BUF, control_genes_file) != NULL) {
    num_genes++;
  }

  /* allocate the memory for the genes */
  *control_genes = malloc(sizeof(unsigned int) * num_genes);

  rewind(control_genes_file);
  while(fgets(line, READ_BUF, control_genes_file) != NULL) {
    /* remove the newline character */
    line[strlen(line) - 1] = '\0';

    /* search for the gene in the annotations */
    for (i = 0; i < num_ann; i++) {
      if (strcmp(line, ann[i].gene_name) == 0) {
	(*control_genes)[count] = i;
      }
    }
    count++;
  }

  free(line);
  fclose(control_genes_file);

  return count;
}

/******************************************************************************/

void recalculate_term_distr(unsigned int *current_term_distr, annotation *ann,
			    unsigned int old_element,
			    unsigned int new_element)
{
  /* update the term distribution by subtracting the annotations of the 
     old element and adding the annotations of the new element */

  unsigned int i;

  /* remove the old annotations */
  for (i = 0; i < ann[old_element].num_terms; i++) {
    current_term_distr[ann[old_element].terms[i]]--;
  }

  /* add the new annotations */
  for (i = 0; i < ann[new_element].num_terms; i++) {
    current_term_distr[ann[new_element].terms[i]]++;
  }
}

/******************************************************************************/

void set_parameters(int argc, char **argv, Parameters *params)
{
  /* get the command line parameters */
  int c;

  /* sanity check */
  if (argc < 17) {
    printf("%s\n", USAGE);
    abort();
  }

  /* set default parameters */
  params->exp_coeff = -1;
  params->max_iter = 1E5;
  params->tolerance = 0;
  params->size_sample = 0;
  params->num_samples = 0;

  /* process the ARGV list */
  while ((c = getopt(argc, argv, "a:c:d:g:m:n:o:s:t:")) != -1) {
    switch (c) {

      case 'a': /* annotations */
	strncpy(params->ann_file_name, optarg, MAX_STR_LEN);
	break;

      case 'c': /* exp coefficient */
	params->exp_coeff = atof(optarg);
        break;

      case 'd': /* target distr */
	strncpy(params->target_file_name, optarg, MAX_STR_LEN);
        break;

      case 'g': /* control genes */
	strncpy(params->control_genes_file_name, optarg, MAX_STR_LEN);
        break;

      case 'm': /* maximum number of steps */
	params->max_iter = atof(optarg);
        break;
    
      case 'n': /* number of samples */
	params->num_samples = atoi(optarg);
        break;

      case 'o': /* output file */
	strncpy(params->output_file_name, optarg, MAX_STR_LEN);
        break;

      case 's': /* sample size */
	params->size_sample = atoi(optarg);
        break;

      case 't': /* tolerance */
	params->tolerance = atof(optarg);
        break;

    default:
      fprintf(stderr, "%s\n", USAGE);
      abort();
    }
  }
}

/******************************************************************************
 * MAIN PROGRAM                                                               *
 ******************************************************************************/

int main(int argc, char **argv)
{
  unsigned int i, num_ann, num_control_genes;
  annotation *ann;
  target_distr target;
  unsigned int *control_genes, *sample;
  FILE *out_file;
  sampling_log log;
  Parameters params;

  /* process the command-line parameters */
  set_parameters(argc, argv, &params);

  /* read the target distribution */
  read_target_distr(params.target_file_name, &target);

  /* read the annotations */
  num_ann = read_annotations(params.ann_file_name, &ann, target);

  /* store the control genes */
  num_control_genes = read_control_genes(params.control_genes_file_name, ann,
					 num_ann, &control_genes);

  /* set up the random number environment */
  gsl_rng_env_setup();
  T = gsl_rng_default;
  rng = gsl_rng_alloc (T);

  /* run the MCMC */
  sample = malloc(sizeof(unsigned int) * params.size_sample);
  out_file = fopen(params.output_file_name, "w");
  for (i = 0; i < params.num_samples; i++) {
    gsl_rng_set(rng, i + 1); /* set the seed */
    log = sampler(target, ann, control_genes, num_control_genes,
		  params.size_sample, params.max_iter,
		  params.tolerance, sample, params.exp_coeff);

    /* print the results */
    print_samples(out_file, sample, params.size_sample, ann);

    /* print a log with the number of iterations required and the squared
       norm */
    printf("%.0f\t%.3f\n", log.num_iter, log.squared_norm);
  }
  fclose(out_file);

  /*******************
   * MEMORY CLEAN UP *
   *******************/

  /* clean up the memory for the target */
  free(target.frequency);
  for (i = 0; i < target.num_terms; i++)
    free(target.terms[i]);
  free(target.terms);

  /* clean up the memory for ann */
  for (i = 0; i < num_ann; i++) {
    free(ann[i].terms);
  }
  free(ann);

  /* clean up the memory for the control genes */
  free(control_genes);

  /* clean up the memory for the sample */
  free(sample);

  return 0;
}
